/*!
 * Ghostery for Opera
 * http://www.ghostery.com/
 *
 * Copyright 2014 EVIDON, Inc. All rights reserved.
 * See https://www.ghostery.com/eula for license.
 */

var Cc = require('chrome').Cc,
	Ci = require('chrome').Ci,
	SDK = {
		system: require('sdk/system')
	},
	conf = require('./conf').load,
	utils = require('./utils');

function getOS() {
	var OS = '';
	switch (SDK.system.platform.toLowerCase()) {
		case 'darwin':
			OS = 'mac';
			break;
		case 'winnt':
			OS = 'win';
			break;
		case 'linux':
			OS = 'linux';
			break;
		case 'openbsd':
			OS = 'openbsd';
			break;
		default:
			OS = 'other';
	}

	return OS;
}

function fetchMktgData(callback) {
	if (!conf.show_cmp) { return; }
	var URL = 'https://cmp-cdn.ghostery.com/check' +
		'?os=' + encodeURIComponent(getOS()) +
		'&gr=' + encodeURIComponent((conf.ghostrank ? 'opt-in' : 'opt-out')) +
		'&ua=firefox' +
		'&v=' + encodeURIComponent(utils.prefs('cmp_version') || 0);

	try {
		var xhr = Cc['@mozilla.org/xmlextras/xmlhttprequest;1'].createInstance(Ci.nsIXMLHttpRequest);

		xhr.open('GET', URL, true);
		xhr.responseType = 'json';
		xhr.onreadystatechange = function () {
			var data;

			if (xhr.readyState == 4 && xhr.status != 204) { // check for success and no content
				data = xhr.response;

				if (data && (!utils.prefs('cmp_version') || data.Version > utils.prefs('cmp_version'))) { // success
					utils.prefs('cmp_version', data.Version);
					// set default dismiss
					data.Campaigns.forEach(function (c) {
						if (c.Dismiss === 0) {
							c.Dismiss = 10;
						}
					});
					utils.prefs('cmp_data', data.Campaigns);
					if (callback) {
						callback();
					}
				} else { // error
					utils.prefs('cmp_data', []);
				}
			}
		};
		xhr.send();
	} catch (e) {
	}
}

exports.fetchMktgData = fetchMktgData;
